import cv2
import numpy as np
from .config import IMAGE_SIZE

def decode_imagebuffer(img_buffer):
    return cv2.imdecode(np.frombuffer(img_buffer, np.uint8), 1)


def processed_image(img):
    """Returns image after preprocessing (undo fisheye, resize)"""
    flat_img = undo_fisheye(img, size=IMAGE_SIZE*2)
    # 480x480 to 224x224
    image = cv2.resize(flat_img, (IMAGE_SIZE, IMAGE_SIZE))
    return image


def encode_image(img):
    """Returns the png encoded array of the input image"""
    #encode to png for compression
    _, encode_array = cv2.imencode('.png', img, [cv2.IMWRITE_PNG_COMPRESSION, 6])
    return encode_array   
 
### fisheye transform

def undo_fisheye(fisheye_img, H=0.5, size=None):
    """Returns the orthographic projection of an fisheye projected img"""
    w = fisheye_img.shape[1]
    #keep size if not specified
    nw = size if size else w
    X, Y = np.ogrid[0:nw, 0:nw]

    #xy distance from center of projection rx,ry in [-1,1]
    rx, ry = np.divide([X, Y], nw/2) - 1

    alpha, epsilon = to_hemishere(rx, ry)
    #angle correction
    epsilon_real = angle_P(epsilon)
    r_proj = to_cloud_plane(epsilon_real, H)
    x_img, y_img = output_coordinate(alpha, r_proj, w)

    return fisheye_img[x_img, y_img]

def angle_P(ideal_angle):
    #Wolkenkamera_FE8174V 4.1 (Korrektur des Höhenwinkels)
    deg = np.rad2deg(ideal_angle)
    ideal_deg = np.polynomial.polynomial.polyval(deg, [2.494295303, 1.326190211, -0.01122405179, 0.0001384399783, -0.0000006380024219])
    return np.deg2rad(ideal_deg)

def angle_Q(real_angle):
    #Wolkenkamera_FE8174V 4.1 (Korrektur des Höhenwinkels)
    deg = np.rad2deg(real_angle)
    real_deg = np.polynomial.polynomial.polyval(deg, [-1.817333483 ,0.6890441144 ,0.008969454015, -0.00009844546547, 0.0000004407488186])
    return np.deg2rad(real_deg)

def to_hemishere(rx, ry):
    #rx, ry normalized radius of source image coords
    r = np.sqrt(np.square(rx) + np.square(ry))
    alpha = np.arctan2(rx,ry)
    epsilon = np.pi/2 * (1 - r)
    return (alpha, epsilon)

def to_cloud_plane(epsilon, H):
    r_real = (1 - epsilon / (np.pi/2))
    eps = np.pi/2 - np.arctan(r_real / H)
    #map to cloud plane with height H
    r_proj = 1 - eps / (np.pi/2)
    return r_proj

def output_coordinate(alpha, radius, size):
    rxp = np.sin(alpha) * radius
    ryp = np.cos(alpha) * radius
    #source coordinates of input image
    return np.around([(rxp + 1)*size/2, (ryp + 1)*size/2]).astype(np.int)