import tensorflow as tf
import sys, os, argparse, zipfile, random

from datetime import timedelta
from concurrent.futures import ThreadPoolExecutor

from .reading import openZips, datetime_to_image, gatherDatapoints
from .image_preproc import processed_image, encode_image, decode_imagebuffer
from .config import IMAGE_DIR, CEILOMETER_DIR, DATASET_DIR, TRAIN_FILE, VALIDATION_FILE, RAIN_ZIP_FILE



def saveTrainAndValidationSet(image_dir=IMAGE_DIR, ceilometer_dir=CEILOMETER_DIR, train=TRAIN_FILE, validation=VALIDATION_FILE, val_ratio=1/5, sample_size=0):
    """Writes a tfrecords training and validation datasets based on data in image_dir and ceilometer_dir"""

    data_points = gatherDatapoints(image_dir, ceilometer_dir)

    pointCount = len(data_points)

    #shuffle data points and sample if requested
    print('Shuffling %d data points.' % pointCount)
    random.shuffle(data_points)
    if sample_size > 0:
        print('using sample size:', sample_size)
        data_points = data_points[:min(len(data_points), sample_size)]
        pointCount = len(data_points)

    # split train and validation by val_ratio
    train_count = int(pointCount * (1- val_ratio))
    train_data_points = data_points[:train_count]
    val_data_points = data_points[train_count:]

    #write train dataset
    trainFilename = os.path.join(DATASET_DIR, train)
    print('Writing training set:', trainFilename)
    saveDataPointsInTFRecords(trainFilename, train_data_points, image_dir)

    #write validation dataset
    valFilename = os.path.join(DATASET_DIR, validation)
    print('Writing validation set:', valFilename)
    saveDataPointsInTFRecords(valFilename, val_data_points, image_dir)


def saveDataPointsInTFRecords(file_name, data_points, image_zip_dir):
    """Writes tfrecord data_set to file_name based on data_points"""
    writer = tf.python_io.TFRecordWriter(file_name)
    writeToTFRecordsFile(writer, data_points, image_zip_dir)
    writer.close()


def make_example(image_zip_dir, ozips, data_point):
    """Returns a tfrecord example based on a data_point after preprocessing the image"""
    date_time, height_reading, rain_reading = data_point
    raw_image = datetime_to_image(date_time, image_zip_dir, ozips)
    image = processed_image(raw_image)

    height, width, depth = image.shape
    image_encoded = encode_image(image)

    return tf.train.Example(features=tf.train.Features(feature={
        'timestamp': _int64_feature(int(date_time.timestamp())),
        'width': _int64_feature(width),
        'height': _int64_feature(height),
        'depth': _int64_feature(depth),
        'height_reading': _float_feature(height_reading),
        'rain_detection': _int64_feature(rain_reading['RD']),
        'rain_quantity': _float_feature(rain_reading['RR']),
        'sun_detection': _int64_feature(rain_reading['GSD']),
        'sun_insolation': _float_feature(rain_reading['G']),
        'image_encoded': _bytes_feature(image_encoded.tostring())}))



def writeToTFRecordsFile(writer, data_points, image_zip_dir):
    """Writes tfrecord examples based on data_points

    Args:
      writer: tfrecord file writer
      data_points: tuples of (date_time, height, rain_dict)
      image_zip_dir: directory of zip files with images
    """

    #use ozips to keep multiple zip files open
    with openZips(10) as ozips:
        pointCount = len(data_points)
        dp_to_example = lambda dp: make_example(image_zip_dir, ozips, dp)

        with ThreadPoolExecutor(max_workers=5) as executor:
            #speed up preprocessing by parallelization
            for i, example in enumerate(executor.map(dp_to_example, data_points)):
                writer.write(example.SerializeToString())
                #inform about progress
                if i % (int(pointCount / 200) + 1) == 0:
                    print('Writing example %d: (%.2f %%)' % (i, 100 * i /pointCount))



def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))

def _float_feature(value):
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))

def _bytes_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

FLAGS = None

def main(_):
    saveTrainAndValidationSet(FLAGS.image_dir, FLAGS.ceilometer_dir,
     FLAGS.train_file, FLAGS.validation_file, FLAGS.val_ratio, FLAGS.sample_size)

if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument('--image_dir', type=str,
                    default=IMAGE_DIR,
                    help='Directory of image zip files')
  parser.add_argument('--ceilometer_dir', type=str,
                    default=CEILOMETER_DIR,
                    help='Directory of ceilometer csv files')
  parser.add_argument('--validation_file', type=str,
                      default=VALIDATION_FILE,
                      help='name of the validation file in out_directory')
  parser.add_argument('--train_file', type=str,
                      default=TRAIN_FILE,
                      help='name of the train file in out_directory')
  parser.add_argument('--sample_size', type=int,
                      default=0,
                      help='only save a sample of all data_points or all with (sample_size = 0)')
  parser.add_argument('--val_ratio', type=float,
                      default=0.2,
                      help='fraction of the data that is saved in the validation dataset')
  FLAGS, unparsed = parser.parse_known_args()
  tf.app.run(main=main, argv=[sys.argv[0]] + unparsed)
