import argparse
import numpy as np
import tensorflow as tf
import os, pickle

from .reading import gatherDatapoints, datetime_to_image, openZips, allRainReadingDetectors, all_ceilometer_readings
from .image_preproc import processed_image, encode_image

from .cloud_cnn import _parse_function, heights_to_labels, num_height_classes, heights_to_labels_np
from .models import ConvnetVGG16

from .config import IMAGE_DIR, CEILOMETER_DIR, DATASET_DIR, RAIN_ZIP_FILE, VALIDATION_FILE, HEIGHT_BINS, SUMMARIES_DIR, BATCH_SIZE

DPS_CACHE_FILE = os.path.join(DATASET_DIR, 'datapoints.pkl')
HEIGHT_CACHE_FILE = os.path.join(DATASET_DIR, 'heightpoints.pkl')

def cached_load_func(gen_fun, cache_file, cache):
    if cache and os.path.isfile(cache_file):
        with open(cache_file, 'rb') as f:
            return pickle.load(f)

    dps = gen_fun()
    if cache:
        with open(cache_file, 'wb') as f:
            pickle.dump(dps, f)
    return dps


def load_datapoints(image_dir=IMAGE_DIR, ceilometer_dir=CEILOMETER_DIR, cache=True):
    gen_fun = lambda: gatherDatapoints(image_dir, ceilometer_dir, fix_infinity=False)
    return cached_load_func(gen_fun, DPS_CACHE_FILE, cache)


def load_heightpoints(ceilometer_dir=CEILOMETER_DIR, cache=True):
    gen_fun = lambda: list(all_ceilometer_readings(ceilometer_dir))
    return cached_load_func(gen_fun, HEIGHT_CACHE_FILE, cache)


def rain_readings():
    for token, detector, start_dt in allRainReadingDetectors(rain_zip_file):
        reg_mask = np.logical_and(detector != 99999, detector > 0)
        reg = detector[reg_mask]
        if token == 'RR':
            rain_quantities = reg
        elif token == 'G':
            sun_insolations = reg

    return (rain_quantities, sun_insolations)


def height_histo(heightpoints, bins=200):
    heights = np.array([h for t,h in heightpoints])

    infinity_count = np.count_nonzero(np.isinf(heights))
    finite_heights = heights[np.isfinite(heights)]
    hist, bins = np.histogram(finite_heights, bins=bins)

    # add inifities to histogram
    hist = np.append(hist, infinity_count)
    bins = np.append(bins, np.inf)

    return hist, bins

def height_pdf(h, hist, bins):
    bin_index = np.where(bins >= h)[-1]
    probs = hist / np.sum(hist)
    return probs[bin_index]

def height_cdf(h, hist, bins):
    bin_indecies = np.where(bins >= h)
    probs = hist / np.sum(hist)
    return np.sum(probs[bin_indecies])

def random_heights(hist, bins, size=1):
    """Genereates numbers acording to the distribution of the histogram"""
    cdf = np.cumsum(hist)
    cdf = cdf / cdf[-1]
    values = np.random.rand(size)
    bin_indecies = np.searchsorted(cdf, values)

    # generate uniform distribution between bin_edges (of finite values)
    fin = bin_indecies + 1 < cdf.size
    finite_indecies = bin_indecies[fin]
    cdf_left = (values[fin] - cdf[finite_indecies])/ (cdf[finite_indecies] - cdf[finite_indecies+1])
    bin_min = bins[finite_indecies]
    bin_max = bins[finite_indecies + 1]

    heights = np.full(bin_indecies.shape, np.inf)
    heights[fin] = np.round(bin_max - cdf_left * (bin_max - bin_min))
    return heights


def test_model(checkpoint_dir=SUMMARIES_DIR, height_regression=False, batch_size=BATCH_SIZE):
    validation_filename = os.path.join(DATASET_DIR, VALIDATION_FILE)
    val_dataset = tf.data.TFRecordDataset(validation_filename).map(
        lambda x: _parse_function(x), num_parallel_calls=2)
    val_dataset = val_dataset.batch(batch_size)
    validation_iterator = val_dataset.make_one_shot_iterator()

    print('Batch size:', batch_size)

    images_big, height_readings, rain_quantity, sun_insolation = validation_iterator.get_next()
    labels = heights_to_labels(height_readings)
    # resize for vgg16 (224x224)
    images = tf.image.resize_images(images_big, (224,224))

    conv_net = ConvnetVGG16(images, classes=num_height_classes(), height_regression=height_regression)
    keep_prob = conv_net.keep_prob

    accuracy = conv_net.accuracy(labels)
    e_loss = conv_net.cross_entropy_loss(labels)
    height_classes = conv_net.class_predictions()
    if height_regression:
        height_predictions = conv_net.height_predictions()
        height_preds = np.array([])
    else: height_preds = None

    batch_accuracies = np.array([])
    height_class_preds = np.array([])
    losses = np.array([])

    height_class_reals = np.array([])
    height_reals = np.array([])

    saver = tf.train.Saver()

    with tf.Session() as sess:
        ckpt = tf.train.get_checkpoint_state(checkpoint_dir)

        if ckpt and ckpt.model_checkpoint_path:
            print('restoring model from:', ckpt.model_checkpoint_path)
            saver.restore(sess, ckpt.model_checkpoint_path)
        else:
            raise ValueError('Could not load checkpoint from {}.'.format(checkpoint_dir))

        while True:
            try:
                extra = [height_preditions] if height_regression else []
                results = sess.run([ accuracy, e_loss, height_classes, labels,
                                   height_readings] + extra, feed_dict={keep_prob: 1.0})
            except (tf.errors.OutOfRangeError, KeyboardInterrupt):
                print('Done testing model.')
                break
            if height_regression:
                val_accuracy, loss, h_classes, h_class_reals, h_reals, h_preds = results
                height_preds = np.append(height_preds, h_preds)
            else:
                val_accuracy, loss, h_classes, h_class_reals, h_reals = results

            batch_accuracies = np.append(batch_accuracies, val_accuracy)
            losses = np.append(losses, loss)
            height_class_preds = np.append(height_class_preds, h_classes)

            height_class_reals = np.append(height_class_reals, h_class_reals)
            height_reals = np.append(height_reals, h_reals)

            print('Processed {} samples. Mean accuracy: {:.2f}, '.format(height_class_reals.size, np.mean(batch_accuracies)))
            if height_regression:
                print('real: {} , prediction:{}'.format(h_reals, h_preds))

    return (height_preds, height_reals, height_class_reals.astype(np.int), height_class_preds.astype(np.int), losses)


def validate_model(ceilometer_dir=CEILOMETER_DIR, checkpoint_dir=SUMMARIES_DIR, height_regression=False):
    height_preds, height_reals, height_class_reals, height_class_preds, losses = \
        test_model(checkpoint_dir, height_regression)

    print_test_result(height_preds, height_reals, height_class_reals, height_class_preds, losses)
    random_model(height_reals, ceilometer_dir)

def random_model(height_reals, ceilometer_dir=CEILOMETER_DIR):
    height_class_reals = heights_to_labels_np(height_reals)
    print('Generating random comparison...')
    hps = load_heightpoints(ceilometer_dir)
    hist, bins = height_histo(hps)
    r_heights_preds = random_heights(hist, bins, size=height_reals.size)

    r_height_class_preds = np.digitize(r_heights_preds, HEIGHT_BINS)
    r_height_class_preds[np.isinf(r_heights_preds)] = len(HEIGHT_BINS) + 1

    print_test_result(r_heights_preds, height_reals, height_class_reals, r_height_class_preds)


def print_test_result(height_preds, height_reals, height_class_reals, height_class_preds, losses=None):
    if height_preds is not None:
        fin_msk = np.isfinite(height_reals)
        print_height_pred_stats(height_preds[fin_msk], height_reals[fin_msk])
    height_counts = np.bincount(height_class_reals)
    correct_counts = np.bincount(height_class_reals[ height_class_reals == height_class_preds])

    total_acc = np.count_nonzero(height_class_reals == height_class_preds)/height_class_reals.size
    total_loss = np.inf
    if losses is not None:
        total_loss = np.mean(losses)
        print('mean loss_acc: {}'.format(np.mean(losses[height_class_reals == height_class_preds])))
    print('Total {} samples. Mean accuracy: {:.2f}, Mean loss {:.2f}, '.format(height_class_reals.size, total_acc, total_loss))
    # bin accuracies 
    for i, b in enumerate(HEIGHT_BINS + [0, np.inf]):
        cc = correct_counts[i] if i in range(len(correct_counts)) else 0
        rc = height_counts[i] if i in range(len(height_counts)) else 0
        bin_desc = '[<{}]'.format(b)
        if b == 0:
            bin_desc = '[>{}]'.format(HEIGHT_BINS[i-1])
        if rc > 0:
            print('{}\t AC {}/{}={:.2f}'.format(bin_desc, cc,rc, cc/rc))
            msk = np.logical_and(height_class_reals == i, np.isfinite(height_reals))
            if rc > 1 and height_preds is not None:
                print_height_pred_stats(height_preds[msk], height_reals[msk])
            if losses is not None:
                print('loss:{:.2f}'.format(np.mean(losses[height_class_reals == i])))
                print('loss_acc:{:.2f}'.format(np.mean(losses[np.logical_and(
                    height_class_reals == i, height_class_preds == height_class_reals)])))
        else:
            print('{}\t\t EMPTY!'.format(bin_desc))


def print_height_pred_stats(height_reals, height_preds):
    if np.std(height_reals) == 0:
        height_correlation = np.corrcoef(height_preds, height_reals)[0,1]
    else: height_correlation = np.inf
    height_abs_error = np.abs(height_preds - height_reals)
    height_mean_error = np.mean(height_abs_error)
    height_std_error = np.std(height_abs_error)

    print('HEIGHT: Corr: {:.2f}, Error: {:.2f} (+- {:.2f})'.format(height_correlation, height_mean_error, height_std_error))


def percentile_height_bins(image_dir=IMAGE_DIR, ceilometer_dir=CEILOMETER_DIR):
    dps = load_datapoints(image_dir, ceilometer_dir)
    heights = np.array([h for t,h,r in dps])
    inf_count = np.count_nonzero(np.isinf(heights))
    real_heights = heights[np.isfinite(heights)]
    print('real: {}, inf: {}'.format(real_heights.size, inf_count))
    bins = np.around(real_heights.size / inf_count)
    print(real_heights.size / inf_count)
    ptgs = np.linspace(0,100, bins + 1)[1:-1]
    return np.percentile(real_heights, ptgs)

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--ceilometer_dir', type=str,
                        default=CEILOMETER_DIR,
                        help='The ceilometer directory')
    parser.add_argument('--image_dir', type=str,
                        default=IMAGE_DIR,
                        help='The cloud image directory')
    parser.add_argument('--height_bins', action='store_true',
                       help='Calculate height bins with percentiles of height histogram')
    parser.add_argument('--random_model', action='store_true',
                       help='Calculate height bin accuracy of random model')
    parser.add_argument('--height_regression', action='store_true',
                       help='Assumes that the model predicts height values (is a height_regression model)')
    parser.add_argument('--checkpoint_dir', type=str,
                        default=SUMMARIES_DIR,
                        help='Directory where checkpoint is stored')

    FLAGS, unparsed = parser.parse_known_args()
    rain_zip_file = os.path.join(DATASET_DIR, RAIN_ZIP_FILE)

    if FLAGS.height_bins:
        print('BINS:\n', percentile_height_bins(FLAGS.image_dir, FLAGS.ceilometer_dir))
        exit()
    if FLAGS.random_model:
        dps = load_datapoints(FLAGS.image_dir, FLAGS.ceilometer_dir)
        heights = np.array([h for t,h,r in dps])
        np.random.shuffle(heights)
        random_model( heights, FLAGS.ceilometer_dir)
        exit()
    validate_model(FLAGS.ceilometer_dir, FLAGS.checkpoint_dir, FLAGS.height_regression)

